@@ -3,27 +3,34 @@ import typing as t
 import keyword
 import logging
 from collections import namedtuple
-from functools import partial
-from prestring import PreString
-from prestring.utils import LazyFormat, LazyArgumentsAndKeywords, LazyKeywords
-from dictknife import deepmerge
 from collections import defaultdict
 from collections import OrderedDict
+from functools import partial
+from prestring import PreString
+from prestring.python import Module
+from prestring.utils import LazyFormat, LazyArgumentsAndKeywords, LazyKeywords
+from dictknife import deepmerge
 from swagger_marshmallow_codegen.langhelpers import (
     LazyCallString,
     titleize,
     clsname_from_path,
 )
+from swagger_marshmallow_codegen.constants import (
+    X_MARSHMALLOW_NAME,
+    X_MARSHMALLOW_INLINE,
+)
 from ..context import Context
 from ..config import ConfigDict
 from .accessor import Accessor
 
 if t.TYPE_CHECKING:
     from swagger_marshmallow_codegen.resolver import Resolver
-
+    from ..context import ContextFactory, InputData, OutputData
 
 logger = logging.getLogger(__name__)
-NAME_MARKER = "x-marshmallow-name"
+NAME_MARKER = X_MARSHMALLOW_NAME
+
+PathInfo = namedtuple("PathInfo", "info, required")
 
 
 class CodegenError(Exception):
@@ -35,7 +42,7 @@ class SchemaWriter:
 
     @classmethod
     def override(cls, *, extra_schema_module=None):
-        return partial(cls, extra_schema_module=extra_schema_module,)
+        return partial(cls, extra_schema_module=extra_schema_module)
 
     def __init__(self, accessor: Accessor, schema_class, *, extra_schema_module=None):
         self.accessor = accessor
@@ -68,11 +75,11 @@ class SchemaWriter:
             field = field["items"]
             if self.resolver.has_ref(field):
                 field_class_name, field = self.resolver.resolve_ref_definition(
-                    d, field, level=1
+                    c, d, field, level=1
                 )
                 # finding original definition
                 if self.resolver.has_ref(field):
-                    ref_name, field = self.resolver.resolve_ref_definition(d, field)
+                    ref_name, field = self.resolver.resolve_ref_definition(c, d, field)
                     if ref_name is None:
                         raise CodegenError("ref: %r is not found", field["$ref"])
 
@@ -108,9 +115,7 @@ class SchemaWriter:
             if caller_name == "fields.Nested":
                 return LazyFormat(
                     "fields.Nested({})",
-                    LazyArgumentsAndKeywords(
-                        [LazyFormat("lambda: {}()", field_class_name)], opts,
-                    ),
+                    LazyArgumentsAndKeywords([c.use_relative(field_class_name)], opts,),
                 )
             else:
                 return LazyFormat("{}({})", caller_name, LazyKeywords(opts))
@@ -120,9 +125,7 @@ class SchemaWriter:
             opts = {k: repr(v) for k, v in opts.items()}
             return LazyFormat(
                 "fields.Nested({})",
-                LazyArgumentsAndKeywords(
-                    [LazyFormat("lambda: {}()", field_class_name)], opts,
-                ),
+                LazyArgumentsAndKeywords([c.use_relative(field_class_name)], opts,),
             )
         elif caller_name == "fields.Dict":
             self.accessor.update_option_on_property(c, field, opts)
@@ -138,11 +141,11 @@ class SchemaWriter:
 
             if self.resolver.has_ref(field):
                 field_class_name, field = self.resolver.resolve_ref_definition(
-                    d, field, level=1
+                    c, d, field, level=1
                 )
                 # finding original definition
                 if self.resolver.has_ref(field):
-                    ref_name, field = self.resolver.resolve_ref_definition(d, field)
+                    ref_name, field = self.resolver.resolve_ref_definition(c, d, field)
                     if ref_name is None:
                         raise CodegenError("ref: %r is not found", field["$ref"])
 
@@ -185,7 +188,7 @@ class SchemaWriter:
         field_class_name = None
         if self.resolver.has_ref(field):
             field_class_name, field = self.resolver.resolve_ref_definition(
-                d, field, level=1
+                c, d, field, level=1
             )
             if self.resolver.has_many(field):
                 return self.write_field_one(
@@ -194,7 +197,7 @@ class SchemaWriter:
 
             # finding original definition
             if self.resolver.has_ref(field):
-                ref_name, field = self.resolver.resolve_ref_definition(d, field)
+                ref_name, field = self.resolver.resolve_ref_definition(c, d, field)
                 if self.resolver.has_many(field):
                     return self.write_field_one(
                         c, d, field_class_name, definition, name, field, opts, many=True
@@ -248,7 +251,7 @@ class SchemaWriter:
         base_classes = [self.schema_class]
         if self.resolver.has_ref(definition):
             ref_name, ref_definition = self.resolver.resolve_ref_definition(
-                d, definition
+                c, d, definition
             )
             if ref_name is None:
                 raise CodegenError("$ref %r is not found", definition["$ref"])
@@ -258,7 +261,7 @@ class SchemaWriter:
                 items = ref_definition["items"]
                 if self.resolver.has_ref(items):
                     _, items = self.resolver.resolve_ref_definition(
-                        d, ref_definition["items"]
+                        c, d, ref_definition["items"]
                     )
                 if not self.resolver.has_schema(d, items):
                     return self.write_primitive_schema(
@@ -276,12 +279,13 @@ class SchemaWriter:
                 base_classes = [ref_name]
         elif self.resolver.has_allof(definition):
             ref_list, ref_definition = self.resolver.resolve_allof_definition(
-                d, definition
+                c, d, definition
             )
             definition = deepmerge(ref_definition, definition)
             if ref_list:
                 base_classes = []
                 for ref_name, ref_definition in ref_list:
+                    c.relative_import(ref_name)
                     if ref_name is None:
                         raise CodegenError(
                             "$ref %r is not found", ref_definition
@@ -309,7 +313,7 @@ class SchemaWriter:
                 return self.write_primitive_schema(c, d, clsname, definition, many=many)
             else:
                 ref_name, ref_definition = self.resolver.resolve_ref_definition(
-                    d, definition["items"]
+                    c, d, definition["items"]
                 )
                 if ref_name is None:
                     return self.write_primitive_schema(
@@ -360,7 +364,7 @@ class SchemaWriter:
                 subdef = definition["additionalProperties"]
                 with c.m.class_("Meta"):
                     if self.resolver.has_ref(subdef):
-                        ref_name, _ = self.resolver.resolve_ref_definition(d, subdef)
+                        ref_name, _ = self.resolver.resolve_ref_definition(c, d, subdef)
                         if ref_name is None:
                             raise CodegenError("$ref %r is not found", subdef["$ref"])
                         self.write_field_one(
@@ -425,7 +429,8 @@ class DefinitionsSchemaWriter:
     def config(self) -> ConfigDict:
         return self.accessor.config
 
-    def write(self, c, d):
+    def write(self, d: InputData, *, context_factory: ContextFactory) -> None:
+        part = self.__class__.__name__
         for schema_name, definition in self.accessor.definitions(d):
             if not self.config.get(
                 "emit_schema_even_primitive_type", False
@@ -433,6 +438,9 @@ class DefinitionsSchemaWriter:
                 logger.info("write schema: skip %s", schema_name)
                 continue
 
+            c = context_factory(
+                definition.get(X_MARSHMALLOW_INLINE) or schema_name, part=part,
+            )
             clsname = self.resolver.resolve_schema_name(schema_name)
             logger.info("write schema: write %s", schema_name)
             self.schema_writer.write_schema(c, d, clsname, definition)
@@ -452,9 +460,10 @@ class PathsSchemaWriter:
     def get_lazy_clsname(self, path):
         return PreString(clsname_from_path(path))
 
-    def write(self, c, d):
+    def write(self, d: InputData, *, context_factory: ContextFactory) -> None:
+        part = self.__class__.__name__
         for path, methods in self.accessor.paths(d):
-            sc = c.new_child()
+            sc = context_factory(path, part=part)
             found = False
             lazy_clsname = self.get_lazy_clsname(path)
             with sc.m.class_(LazyFormat("{}Input", lazy_clsname)):
@@ -481,7 +490,10 @@ class PathsSchemaWriter:
                             ssc.m.stmt("")
 
                         path_info = self.build_path_info(
-                            d, toplevel_parameters, self.accessor.parameters(definition)
+                            sc,
+                            d,
+                            toplevel_parameters,
+                            self.accessor.parameters(definition),
                         )
                         for section, properties in sorted(path_info.info.items()):
                             if section is None:
@@ -509,21 +521,24 @@ class PathsSchemaWriter:
             if not found:
                 sc.m.clear()
 
-    PathInfo = namedtuple("PathInfo", "info, required")
-
-    def build_path_info(self, fulldata, *paramaters_set):
+    def build_path_info(
+        self,
+        c: Context,
+        fulldata: t.Dict[str, t.Any],
+        *paramaters_set: t.List[t.Dict[str, t.Any]]
+    ) -> PathInfo:
         info = defaultdict(OrderedDict)
         required = defaultdict(list)
         for parameters in paramaters_set:
             for p in parameters:
                 if self.resolver.has_ref(p):
-                    _, p = self.resolver.resolve_ref_definition(fulldata, p)
+                    _, p = self.resolver.resolve_ref_definition(c, fulldata, p)
                 name = p.get("name")
                 section = p.get("in")
                 info[section][name] = p
                 if p.get("required"):
                     required[section].append(name)
-        return self.PathInfo(info=info, required=required)
+        return PathInfo(info=info, required=required)
 
 
 class ResponsesSchemaWriter:
@@ -541,10 +556,11 @@ class ResponsesSchemaWriter:
     def get_lazy_clsname(self, path):
         return PreString(clsname_from_path(path))
 
-    def write(self, c, d):
+    def write(self, d: InputData, *, context_factory: ContextFactory) -> None:
+        part = self.__class__.__name__
         for path, methods in self.accessor.paths(d):
             lazy_clsname = self.get_lazy_clsname(path)
-            sc = c.new_child()
+            sc = context_factory(path, part=part)
             found = False
             with sc.m.class_(LazyFormat("{}Output", lazy_clsname)):
                 if self.OVERRIDE_NAME_MARKER in methods:
@@ -554,7 +570,7 @@ class ResponsesSchemaWriter:
                     for status, definition in self.accessor.responses(definition):
                         if self.resolver.has_ref(definition):
                             _, definition = self.resolver.resolve_ref_definition(
-                                d, definition
+                                sc, d, definition
                             )
                         if "schema" in definition:
                             found = True
@@ -589,41 +605,54 @@ class Codegen:
         )
         self.schema_class = self.schema_class_path.rsplit(":", 1)[-1]
 
+        self.files: t.Dict[str, Module] = {}
+
     @property
     def resolver(self) -> Resolver:
         return self.accessor.resolver
 
-    def write_header(self, c, *, comment: t.Optional[str] = None):
+    def write_header(self, c: Context, *, comment: t.Optional[str] = None) -> None:
         if comment is None:
             comment = """\
 # this is auto-generated by swagger-marshmallow-codegen
 from __future__ import annotations
 """
+        if not comment:
+            return
+
         for line in comment.splitlines():
             c.im.stmt(line)
 
-    def write_import_(self, c):
+    def write_import_(self, c: Context) -> None:
         c.from_(*self.schema_class_path.rsplit(":", 1))
         c.from_("marshmallow", "fields")
 
-    def write_body(self, c, d):
+    def write_body(self, d: InputData, *, context_factory: ContextFactory) -> None:
+        # TODO: get from context
         sw = self.schema_writer_factory(self.accessor, self.schema_class)
-        config = self.accessor.config
-        if config.get("schema", False):
-            DefinitionsSchemaWriter(self.accessor, sw).write(c.new_child(), d)
-        if config.get("input", False):
-            PathsSchemaWriter(self.accessor, sw).write(c.new_child(), d)
-        if config.get("output", False):
-            ResponsesSchemaWriter(self.accessor, sw).write(c.new_child(), d)
 
-    def codegen(self, d, ctx=None):
-        c = ctx or Context()
-        if not self.accessor.config.get("skip_header_comment", False):
-            self.write_header(c, comment=self.accessor.config.get("header_comment"))
-        c.m.sep()
-        self.write_import_(c)
-        self.write_body(c, d)
-        return c.m
+        config = self.accessor.config
+        if config.get("emit_schema", False):
+            DefinitionsSchemaWriter(self.accessor, sw).write(
+                d, context_factory=context_factory
+            )
+        if config.get("emit_input", False):
+            PathsSchemaWriter(self.accessor, sw).write(
+                d, context_factory=context_factory
+            )
+        if config.get("emit_output", False):
+            ResponsesSchemaWriter(self.accessor, sw).write(
+                d, context_factory=context_factory
+            )
+
+    def setup_context(self, ctx: Context) -> None:
+        self.write_header(ctx, comment=self.accessor.config.get("header_comment"))
+        self.write_import_(ctx)
+        ctx.m.sep()
+
+    def codegen(self, d: InputData, context_factory) -> OutputData:
+        self.write_body(d, context_factory=context_factory)
+        return context_factory
 
 
 def lazy_json_dump(s):
